from functools import partial, wraps
from importlib import import_module, invalidate_caches as importlib_invalidate_caches
from logging import INFO
from pkgutil import walk_packages
from typing import Any, Callable

from click import (
    ClickException,
    group,
    option as click_option,
    Path as ClickPath,
)
from coloredlogs import install as coloredlogs_install

from invoicez.exceptions import InvoicezException


option = partial(click_option, show_default=True)

dir_path_option = option(
    "--dir-path",
    type=ClickPath(exists=True, readable=True, file_okay=False),
    default=".",
    help="Path of the deck.",
)


@group(chain=True)
def cli() -> None:
    coloredlogs_install(
        level=INFO, fmt="%(asctime)s %(name)s %(message)s", datefmt="%H:%M:%S",
    )


def command(f: Callable[..., Any]) -> Callable[..., Any]:
    @wraps(f)
    def wrapper(*args: Any, **kwargs: Any) -> Any:
        try:
            result = f(*args, **kwargs)
        except InvoicezException as e:
            raise ClickException(str(e)) from e
        return result

    return cli.command()(wrapper)


def _import_module_and_submodules(package_name: str) -> None:
    """
    From https://github.com/allenai/allennlp/blob/master/allennlp/common/util.py
    """
    importlib_invalidate_caches()

    module = import_module(package_name)
    path = getattr(module, "__path__", [])
    path_string = "" if not path else path[0]

    for module_finder, name, _ in walk_packages(path):
        if path_string and module_finder.path != path_string:
            continue
        subpackage = f"{package_name}.{name}"
        _import_module_and_submodules(subpackage)


_import_module_and_submodules(__name__)



from filecmp import cmp
from logging import getLogger
from multiprocessing import Pool
from os import unlink
from os.path import join as path_join
from pathlib import Path
from shutil import copyfile, move
from subprocess import CompletedProcess, run
from tempfile import NamedTemporaryFile
from typing import Any, Dict, Optional

from jinja2 import Environment, FileSystemLoader

from invoicez.exceptions import InvoicezException
from invoicez.paths import Paths
from invoicez.target import Target


class Builder:
    def __init__(self, target: Target, config: Dict[str, Any], paths: Paths):
        self._target = target
        self._config = config
        self._paths = paths
        self._logger = getLogger(__name__)
        self._track_build()

    def _track_build(self) -> None:
        to_compile = [self._target]
        with Pool(1) as pool:
            completed_processes = pool.map(self._build, to_compile)
        for completed_process, target in zip(completed_processes, to_compile):
            compilation = "%s/%s" % (target.name, target.template_name)
            if completed_process.returncode != 0:
                self._logger.warning("Compilation %s errored", compilation)
                self._logger.warning(
                    "Captured %s stderr\n%s", compilation, completed_process.stderr
                )
                self._logger.warning(
                    "Captured %s stdout\n%s", compilation, completed_process.stdout
                )

    def _get_filename(self) -> str:
        return self._target.name

    def _build(
        self, target: Optional[Target], copy_result: bool = True
    ) -> CompletedProcess:
        build_dir = self._setup_build_dir()
        filename = self._get_filename()
        latex_path = build_dir / f"{filename}.tex"
        build_pdf_path = latex_path.with_suffix(".pdf")
        output_pdf_path = self._paths.pdf_dir / f"{filename}.pdf"

        self._write_latex(latex_path)

        completed_process = self._compile(
            latex_path=latex_path.relative_to(build_dir), build_dir=build_dir,
        )
        if copy_result and completed_process.returncode == 0:
            self._paths.pdf_dir.mkdir(parents=True, exist_ok=True)
            copyfile(build_pdf_path, output_pdf_path)
        return completed_process

    def _setup_build_dir(self) -> Path:
        target_build_dir = (
            self._paths.build_dir / self._target.name / self._target.template_name
        )
        target_build_dir.mkdir(parents=True, exist_ok=True)
        if self._paths.assets_dir.is_dir():
            for item in self._paths.assets_dir.iterdir():
                self._setup_link(target_build_dir / item.name, item)
        return target_build_dir

    def _write_latex(self, output_path: Path) -> None:
        template = self._env.get_template(str(self._target.template_path.name))
        try:
            with NamedTemporaryFile("w", encoding="utf8", delete=False) as fh:
                fh.write(
                    template.render(
                        **self._config, **self._target.data, target=self._target
                    )
                )
                fh.write("\n")
            if not output_path.exists() or not cmp(fh.name, str(output_path)):
                move(fh.name, output_path)
        finally:
            try:
                unlink(fh.name)
            except FileNotFoundError:
                pass

    @property
    def _env(self) -> Environment:
        if not hasattr(self, "__env"):
            self.__env = Environment(
                loader=FileSystemLoader(searchpath=self._paths.jinja2_dir),
                block_start_string=r"\BLOCK{",
                block_end_string="}",
                variable_start_string=r"\V{",
                variable_end_string="}",
                comment_start_string=r"\#{",
                comment_end_string="}",
                line_statement_prefix="%%",
                line_comment_prefix="%#",
                trim_blocks=True,
                autoescape=False,
            )
            self.__env.filters["camelcase"] = _to_camel_case
            self.__env.filters["path_join"] = lambda paths: path_join(*paths)
        return self.__env

    @staticmethod
    def _compile(latex_path: Path, build_dir: Path) -> CompletedProcess:
        command = [
            "latexmk",
            "-pdflatex=xelatex -shell-escape -interaction=nonstopmode %O %S",
            "-dvi-",
            "-ps-",
            "-pdf",
        ]
        command.append(str(latex_path))
        return run(command, cwd=build_dir, capture_output=True, encoding="utf8")

    @staticmethod
    def _setup_link(source: Path, target: Path) -> None:
        if not target.exists():
            raise InvoicezException(
                f"{target} could not be found. Please make sure it exists before "
                "proceeding."
            )
        target = target.resolve()
        if source.is_symlink():
            if source.resolve().samefile(target):
                return
            raise InvoicezException(
                f"{source} already exists in the build directory and does not point to "
                f"{target}. Please clean the build directory."
            )
        elif source.exists():
            raise InvoicezException(
                f"{source} already exists in the build directory. Please clean the "
                "build directory."
            )
        source.parent.mkdir(parents=True, exist_ok=True)
        source.symlink_to(target)


def _to_camel_case(string: str) -> str:
    return "".join(substring.capitalize() or "_" for substring in string.split("_"))



from logging import getLogger
from pathlib import Path
from shutil import rmtree

from invoicez.cli import command, dir_path_option
from invoicez.paths import Paths


@command
@dir_path_option
def clean(dir_path: str) -> None:
    """Wipe the build directory."""
    logger = getLogger(__name__)
    paths = Paths(Path(dir_path))
    if not paths.build_dir.exists():
        logger.info(f"Nothing to do: {paths.build_dir} doesn't exist")
    else:
        logger.info(f"Deleting {paths.build_dir}")
        rmtree(paths.build_dir)



from logging import getLogger
from pathlib import Path

from yaml import safe_load as yaml_safe_load

from invoicez.exceptions import InvoicezException
from invoicez.paths import Paths


_logger = getLogger(__name__)


class Target:
    def __init__(self, path: Path, template_name: str, paths: Paths):
        if not path.is_file():
            raise InvoicezException(f"Could not find the file to compile {path}.")
        self.data = dict(invoice_number=path.with_suffix("").name)
        with path.open(encoding="utf8") as fh:
            self.data.update(yaml_safe_load(fh))
        self.name = path.with_suffix("").name
        self.path = path
        self.template_name = template_name
        self.template_path = paths.jinja2_dir / template_name
        if not self.template_path.name.endswith(".tex"):
            self.template_path = self.template_path.with_suffix(".tex")
        if not self.template_path.is_file():
            raise InvoicezException(
                f"Could not find the template to use {self.template_path}."
            )



from datetime import datetime, timedelta
from logging import getLogger
from pathlib import Path

from invoicez.cli import command, dir_path_option
from invoicez.paths import Paths


_logger = getLogger(__name__)


@command
@dir_path_option
def new(dir_path: str) -> None:
    """Create a new invoice."""
    paths = Paths(Path(dir_path))
    now = datetime.now()
    prefix = now.strftime("%Y-%m")
    date = now.strftime("%d/%m/%Y")
    limit_date = (now + timedelta(days=31)).strftime("%d/%m/%Y")
    n = len(list(paths.git_dir.glob(f"*/{prefix}-*.yml"))) + 1
    output_path = paths.working_dir / f"{prefix}-{n:03}.yml"
    content = paths.template_invoice.read_text(encoding="utf8")
    content = content.replace("{DATE}", date)
    content = content.replace("{LIMIT_DATE}", limit_date)
    _logger.info(
        f"Copying invoice template to {output_path.relative_to(paths.working_dir)}"
    )
    output_path.write_text(content, encoding="utf8")



from pathlib import Path

from invoicez.builder import Builder
from invoicez.config import get_config
from invoicez.paths import Paths
from invoicez.target import Target


def run(path: Path, template: str, paths: Paths) -> None:
    config = get_config(paths)
    target = Target(path, template, paths)
    Builder(target, config, paths)



from collections import ChainMap, OrderedDict
from logging import getLogger
from pathlib import Path
from shutil import copy as shutil_copy
from typing import Any, Dict, Optional

from yaml import safe_load as yaml_safe_load

from invoicez.exceptions import InvoicezException
from invoicez.paths import Paths


_logger = getLogger(__name__)


def get_config(paths: Paths) -> Dict[str, Any]:
    return OrderedDict(
        (k, v)
        for k, v in sorted(
            ChainMap(
                *[
                    _get_or_create_config(f, p)
                    for f, p in [
                        (paths.company_config, paths.template_company_config),
                        (paths.global_config, None),
                    ]
                ],
            ).items()
        )
    )


def _get_or_create_config(path: Path, template_path: Optional[Path]) -> Dict[str, Any]:
    if not path.is_file():
        if template_path:
            if template_path.is_file():
                shutil_copy(str(template_path), str(path), follow_symlinks=True)
                raise InvoicezException(
                    f"{path} was not found, copied {template_path} there. Please edit "
                    "it."
                )
            else:
                raise InvoicezException(
                    f"Neither {path} nor {template_path} were found. "
                    "Please create both."
                )
        else:
            raise InvoicezException(f"{path} was not found. Please create it.")
    with open(path, "r", encoding="utf8") as fh:
        return yaml_safe_load(fh)



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['invoicez', 'invoicez.cli']

package_data = \
{'': ['*']}

install_requires = \
['GitPython>=3.1.0,<4.0.0',
 'Jinja2>=2.11.1,<3.0.0',
 'PyYAML>=5.3.1,<6.0.0',
 'click>=7.1.1,<8.0.0',
 'coloredlogs>=14.0,<15.0']

entry_points = \
{'console_scripts': ['invoicez = invoicez.cli:cli']}

setup_kwargs = {
    'name': 'invoicez',
    'version': '1.0.1',
    'description': 'Tool to handle invoices.',
    'long_description': "# `invoicez`\n\n[![CI Status](https://img.shields.io/github/workflow/status/m09/invoicez/CI?label=CI&style=for-the-badge)](https://github.com/m09/invoicez/actions?query=workflow%3ACI)\n[![CD Status](https://img.shields.io/github/workflow/status/m09/invoicez/CD?label=CD&style=for-the-badge)](https://github.com/m09/invoicez/actions?query=workflow%3ACD)\n[![Test Coverage](https://img.shields.io/codecov/c/github/m09/invoicez?style=for-the-badge)](https://codecov.io/gh/m09/invoicez)\n[![PyPI Project](https://img.shields.io/pypi/v/invoicez?style=for-the-badge)](https://pypi.org/project/invoicez/)\n\nTool to handle invoices. It is currently not meant to be usable directly by people finding about the package on GitHub. Please open an issue if you want more details or want to discuss this solution.\n\n## Installation\n\nWith `pip`:\n\n    pip install invoicez\n\n## Directory Structure\n\n`invoicez` works with big assumptions on the directory structure of your presentation repository. Among those assumptions:\n\n- your directory should be a git repository\n- it should contain a jinja2 LaTeX invoice template in the `jinja2` directory, with a specific name (`main.tex.jinja2`)\n- your invoice folders should be contained in an organization/company folder. This is meant to avoid repeating the company details all over the place\n\n```\nroot (git repository)\n├── global-config.yml\n├── jinja2\n│\xa0\xa0 ├── main.tex.jinja2\n├── assets\n│\xa0\xa0 ├── img\n│\xa0\xa0 │\xa0\xa0 ├── logo.png\n│\xa0\xa0 │\xa0\xa0 └── signature.jpg\n├── company1\n│\xa0\xa0 └── company-config.yml\n│\xa0\xa0  \xa0\xa0 ├── invoice1.yml\n│\xa0\xa0  \xa0\xa0 └── invoice2.yml\n└── company2\n \xa0\xa0 └── company-config.yml\n \xa0\xa0  \xa0\xa0 ├── invoice1.yml\n \xa0\xa0  \xa0\xa0 └── invoice2.yml\n```\n\n## Configuration\n\n`invoicez` uses two configuration files to avoid repetition, one for your details and global values, one for the details of the company you're writing the invoice for.\n\n### Configuration merging\n\nThe company config is merged into the global config, you can use that fact to override global values for a specific company.\n\n### Using the configuration values\n\nThe values obtained from the merged configurations can be used directly in Jinja2 templates, or in LaTeX after a conversion from snake case to camel case: if the configuration contains the key `user_email`, it will be defined as the `\\UserEmail` command in LaTeX.\n\n## Usage\n\nSee the `--help` flag of the `invoicez` command line tool.\n",
    'author': 'm09',
    'author_email': '142691+m09@users.noreply.github.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/m09/invoicez',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.6,<4.0',
}


setup(**setup_kwargs)



from logging import getLogger
from pathlib import Path
from typing import Optional

from git import Repo
from git.exc import InvalidGitRepositoryError

from invoicez.exceptions import InvoicezException


_logger = getLogger(__name__)


class Paths:
    def __init__(self, working_dir: Path, check_depth: bool = True) -> None:
        self.working_dir = working_dir.resolve()
        self.build_dir = self.working_dir / "build"
        self.pdf_dir = self.working_dir / "pdf"
        self.assets_dir = self.git_dir / "assets"
        self.templates_dir = self.git_dir / "templates"
        self.yml_templates_dir = self.templates_dir / "yml"
        self.template_company_config = self.yml_templates_dir / "company-config.yml"
        self.template_invoice = self.yml_templates_dir / "invoice.yml"
        self.jinja2_dir = self.templates_dir / "jinja2"
        self.global_config = self.git_dir / "global-config.yml"
        self.company_config = (
            self.git_dir
            / self.working_dir.relative_to(self.git_dir).parts[0]
            / "company-config.yml"
        )

        if check_depth and not self.working_dir.relative_to(self.git_dir).match("*"):
            raise InvoicezException(
                f"Not deep enough from root {self.git_dir}. "
                "Please follow the directory hierarchy root > company and "
                "invoke this tool from the company directory."
            )

    @property
    def git_dir(self) -> Optional[Path]:
        if not hasattr(self, "_git_dir"):
            try:
                repository = Repo(str(self.working_dir), search_parent_directories=True)
            except InvalidGitRepositoryError as e:
                raise InvoicezException(
                    "Could not find the path of the current git working directory. "
                    "Are you in one?"
                ) from e
            self._git_dir = Path(repository.git.rev_parse("--show-toplevel")).resolve()
        return self._git_dir



from pathlib import Path
from typing import Any, List

from click import argument, Path as ClickPath

from invoicez.cli import command, dir_path_option, option
from invoicez.paths import Paths
from invoicez.runner import run as runner_run


def _autocomplete_path(ctx: Any, args: List[str], incomplete: str) -> List[str]:
    try:
        paths = Paths(Path("."))
        return [
            str(t.relative_to(paths.working_dir))
            for t in paths.working_dir.glob("*.yml")
            if t.name != "company-config.yml"
        ]
    except Exception:
        return []


def _autocomplete_template(ctx: Any, args: List[str], incomplete: str) -> List[str]:
    try:
        paths = Paths(Path("."))
        return [
            t.with_suffix("").with_suffix("").name
            for t in paths.jinja2_dir.glob("*.tex.jinja2")
        ]
    except Exception:
        return []


@command
@argument(
    "path",
    type=ClickPath(exists=True, dir_okay=False, readable=True),
    autocompletion=_autocomplete_path,  # type: ignore
)
@option(
    "--template",
    type=str,
    default="main",
    autocompletion=_autocomplete_template,  # type: ignore
)
@dir_path_option
def run(path: str, template: str, dir_path: str) -> None:
    """Run the compiler."""
    paths = Paths(Path(dir_path))
    runner_run(Path(path), template, paths)
